/*
  eigen_inline.h

  ͭͷ׻ѥ饤󡦥ƥץ졼ȴؿ
*/

namespace tadah_fussy
{
  namespace math_lib
  {
    namespace eigen_lib
    {
      /**
	 ࡦߥåȤľ
      **/

      /*
	GramSchmidt : ࡦߥåȤľ򲽤ˤäƹ a ٥ȥľѴ

	ͤȤƹ a Υ󥯤֤

	e : ΥबǤ뤫ɤȽꤹ뤿Τ
      */
      template< class T >
      typename SquareMatrix< T >::size_type
      GramSchmidt( SquareMatrix< T >* a, T e )
      {
	using size_type = typename SquareMatrix< T >::size_type;           // Υη

	const std::string FUNC_NAME = "EigenLib::GramSchmidt";

	assert( a != nullptr );
	error_lib::NNNum( e, FUNC_NAME + " : e " + error_message::NNNUM );

	size_type rank = 0;          // 
	size_type sz = a->size();    // Υ
	std::vector< T > norm( sz ); // ٥ȥΥΥ

	auto col = a->column( 0 );
	norm[0] = std::inner_product( col, col.end(), col, T() );
	if ( norm[0] >= e ) ++rank;

	for ( size_type n = 1 ; n < sz ; ++n ) {
	  auto col1 = a->column( n );
	  /* Ѥη׻ */
	  std::vector< T > inPro( n ); // 
	  for ( size_type c = 0 ; c < n ; ++c ) {
	    auto col2 = a->column( c );
	    inPro[c] = std::inner_product( col1, col1.end(), col2, T() );
	  }
	  /* ׻ľ */
	  for ( size_type c = 0 ; c < n ; ++c ) {
	    if ( norm[c] == 0 ) continue;
	    for ( size_type r = 0 ; r < sz ; ++r )
	      (*a)[r][n] -= inPro[c] * (*a)[r][c] / norm[c];
	  }
	  /* Υη׻ */
	  auto col = a->column( n );
	  norm[n] = std::inner_product( col, col.end(), col, T() );
	  if ( norm[n] >= e ) ++rank;
	}

	return( rank );
      }

      /**
	 䥳ˡ
      **/

      /*
	GetMaxValue : гʬͤͤ

	a : оݤоι
	maxR, maxC : ͤäǤιֹѿؤΥݥ
      */
      template< class T >
      void GetMaxValue
      ( const SymmetricMatrix< T >& a,
	typename SymmetricMatrix< T >::size_type* maxR, typename SymmetricMatrix< T >::size_type* maxC )
      {
	using size_type = typename SymmetricMatrix< T >::size_type; // ΥФ뷿

	size_type sz = a.size(); // Υ

	if ( sz == 1 ) {
	  *maxR = *maxC = 0;
	  return;
	}

	*maxR = 0;
	*maxC = 1;

	// ͤǽгʬǽ
	T maxValue = std::abs( a[*maxR][*maxC] );

	for ( size_type r = 0 ; r < sz ; ++r ) {
	  for ( size_type c = r + 1 ; c < sz ; ++c ) {
	    T t = std::abs( a[r][c] );
	    if ( maxValue < t ) {
	      *maxR = r;
	      *maxC = c;
	      maxValue = t;
	    }
	  }
	}
      }

      /*
	GivensRotation_ST : ֥󥹲žˤѴ(гʬΤ⤰á)

	mat : оݤιؤΥݥ
	i, j : õоΤȤʤǤιֹ
	g : 
	sin, cos : sin, cos
	sc : sincos
      */
      template< class T >
      void GivensRotation_ST
      ( SymmetricMatrix< T >* mat,
	typename SymmetricMatrix< T >::size_type i, typename SymmetricMatrix< T >::size_type j,
	T g, T sin, T cos, T sc )
      {
	using size_type = typename SymmetricMatrix< T >::size_type;

	// гʬ (i,i),(j,j)
	T d1 = ( (*mat)[i][i] * ( 1 + g ) + (*mat)[j][j] * ( 1 - g ) ) / 2 - 2 * (*mat)[i][j] * sc;
	T d2 = ( (*mat)[j][j] * ( 1 + g ) + (*mat)[i][i] * ( 1 - g ) ) / 2 + 2 * (*mat)[i][j] * sc;
	// гʬ (i,j),(j,i)
	T d3 = ( (*mat)[i][j] * g ) + ( (*mat)[i][i] - (*mat)[j][j] ) * sc;

	(*mat)[i][i] = d1;
	(*mat)[j][j] = d2;
	(*mat)[i][j] = (*mat)[j][i] = d3;

	// ¾гʬ
	for ( size_type r = 0 ; r < mat->size() ; ++r ) {
	  if ( r == i || r == j ) continue;
	  d1 = (*mat)[r][i] * cos - (*mat)[r][j] * sin;
	  d2 = (*mat)[r][i] * sin + (*mat)[r][j] * cos;
	  (*mat)[r][i] = d1;
	  (*mat)[r][j] = d2;
	}
      }

      /*
	GivensRotationMul_R : ֥󥹲žѹѤη׻(¦ݤ)

	eigen : оݤιؤΥݥ
	i, j : õоݤȤʤǤιֹ
	sin, cos : sin,cos
      */
      template< class T >
      void GivensRotationMul_R
      ( SquareMatrix< T >* eigen,
	typename SquareMatrix< T >::size_type i, typename SquareMatrix< T >::size_type j,
	T sin, T cos )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	for ( size_type r = 0 ; r < eigen->size() ; ++r ) {
	  T d1 = (*eigen)[r][i] * cos - (*eigen)[r][j] * sin;
	  T d2 = (*eigen)[r][i] * sin + (*eigen)[r][j] * cos;
	  (*eigen)[r][i] = d1;
	  (*eigen)[r][j] = d2;
	}
      }

      /*
	Jacobi : 䥳ˡˤ mat θͭͻ

	᤿ͭͤгʬȤƻĤ롣

	eigen ͭ٥ȥ뤿ιؤΥݥ(NULLʤ̵)
	e «Ƚꤹ뤿Τ
	maxCnt «ʤäκ׻
      */
      template< class T >
      void Jacobi( SymmetricMatrix< T >* mat, SquareMatrix< T >* eigen, T e, unsigned int maxCnt )
      {
	using size_type = typename SymmetricMatrix< T >::size_type;

	const std::string FUNC_NAME = "EigenLib::Jacobi";

	assert( mat != nullptr );
	error_lib::NNNum( e, FUNC_NAME + " : e " + error_message::NNNUM );

	unsigned int cnt = 0; // 

	if ( eigen != nullptr ) {
	  eigen->assign( mat->size() );
	  for ( size_type i = 0 ; i < mat->size() ; ++i )
	    (*eigen)[i][i] = 1;
	}

	while ( cnt < maxCnt ) {
	  size_type i, j;
	  GetMaxValue( *mat, &i, &j );
	  if ( std::abs( (*mat)[i][j] ) < e ) break;
	  T a = ( (*mat)[i][i] - (*mat)[j][j] ) / 2;        //  : i, j ܤгʬκȾʬ
	  T b = (*mat)[i][j] * -1;                          //  : гʬȿž
	  T g = std::abs( a ) / std::sqrt( a * a + b * b ); //  : |a| / d( a, b )
	  T sin = std::sqrt( ( 1 - g ) / 2 );               // sin
	  T cos = std::sqrt( ( 1 + g ) / 2 );               // cos
	  T sc = std::sqrt( 1 - g * g ) / 2;                // sincos
	  if ( a * b < 0 ) {                                // sign()Υå
	    sin *= -1;
	    sc *= -1;
	  }
	  GivensRotation_ST( mat, i, j, g, sin, cos, sc ); // ֥󥹲žˤѴ(гʬΤ⤰á)

	  // ͭ٥ȥη׻
	  if ( eigen != nullptr ) {
	    if ( cnt == 0 ) {
	      (*eigen)[i][i] = (*eigen)[j][j] = cos;
	      (*eigen)[i][j] = sin;
	      (*eigen)[j][i] = -sin;
	    } else {
	      GivensRotationMul_R( eigen, i, j, sin, cos );
	    }
	  }

	  ++cnt;
	}

#ifdef DEBUG
	std::cout << FUNC_NAME + " : cnt = " << cnt << std::endl;
#endif

	// Ķ㳰ꤲ
	if ( cnt >= maxCnt )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt );
      }

      /**
	 ϥۥˡ
      **/

      /*
	CreateVec : ǡ x ζѴ(ϥۥѴ)ԤѴΥ٥ȥΥΥʿ֤

	٥ȥ x ȥΥबʬʳˤʤ褦ʥ٥ȥ x' ᡢ
	v = x - x' Ѵ(ϥۥѴ)Υ ||x|| ( = ||x'|| ) ֤ͤȤ֤
      */
      template< class T >
      T CreateVec( std::vector< T >* x )
      {
	T v = std::sqrt( std::inner_product( x->begin(), x->end(), x->begin(), T() ) ); // vec ΥΥ

	if ( x->front() > 0 )
	  x->front() += v;
	else
	  x->front() -= v;

	return( std::inner_product( x->begin(), x->end(), x->begin(), T() ) );
      }

      /*
	MultHouseholder :  mat ˥ϥۥݤ

	H(v) = E - 2(v)(vT) / ||v||

	E : ñ̹
	v : Ѵ٥ȥ , vT : vž֥٥ȥ , ||v|| : vΥΥ

	 A ФH(v)A ׻

	оϰϤ col ʹߤιԤȤ(ǤѴԤ)
	norm  v ΥΥǤ롣
      */
      template< class T >
      void MultHouseholder
      ( SquareMatrix< T >* mat, typename SquareMatrix< T >::size_type col,
	const std::vector< T >& v, T norm )
      {
	using sm_size = typename SquareMatrix< T >::size_type;
	using vec_size = typename std::vector< T >::size_type;

	for ( sm_size c = 0 ; c < mat->cols() ; ++c ) {
	  // ׻оݤ٥ȥǤХåե˥ԡ
	  auto colIt = mat->column( c, col );
	  std::vector< T > buff( colIt, colIt.end() );

	  // Ǥ˶ѴԤϽ
	  for ( sm_size r = col ; r < mat->rows() ; ++r ) {
	    T d = 0;
	    for ( vec_size i = 0 ; i < v.size() ; ++i )
	      d -= v[i] * buff[i];
	    d *= 2 * v[r - col] / norm;
	    (*mat)[r][c] = buff[r - col] + d;
	  }
	}
      }

      /*
	Householder_QR : ϥۥˡˤ r  QR Ѵ

	r Ѵ(̤Ͼ廰ѹˤʤ)
	q ľ뤿ι(NULLʤ̵)
      */
      template< class T >
      void Householder_QR( SquareMatrix< T >* r, SquareMatrix< T >* q )
      {
	assert( r != nullptr );

	using size_type = typename SquareMatrix< T >::size_type;

	if ( q != nullptr )
	  q->assign( r->rows() );

	for ( size_type c = 0 ; c < r->cols() - 1 ; ++c ) {
	  // оݤ٥ȥѴ
	  auto colIt = r->column( c, c );
	  std::vector< T > v( colIt, colIt.end() );
	  T norm = CreateVec( &v );

	  MultHouseholder( r, c, v, norm );

	  if ( q == nullptr ) continue;

	  // ľη׻
	  if ( c == 0 ) {
	    for ( size_type j = 0 ; j < r->cols() ; ++j ) {
	      (*q)[j][j] = 1 - 2 * v[j] * v[j] / norm;
	      for ( size_type i = j + 1 ; i < r->rows() ; ++i )
		(*q)[j][i] = (*q)[i][j] = - 2 * v[j] * v[i] / norm;
	    }
	  } else {
	    MultHouseholder( q, c, v, norm );
	  }
	}

	if ( q != nullptr )
	  q->transpose(); // ľž
      }

      /**
	 ϥۥˡˤ뻰гѹ󲽤 QR Ѵ
      **/

      /*
	Householder_ST : ϥۥˤ mat Ѵ

	v ϶Ѵ٥ȥǡϥۥǤȤʤ롣
	оϰϤ col ʹߤιԤȤ(ǤѴԤ)
	norm  v ΥΥǤ롣
      */
      template< class T >
      void Householder_ST
      ( SquareMatrix< T >* mat, typename SquareMatrix< T >::size_type col,
	const std::vector< T >& v, T norm )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	SquareMatrix< T > buff( mat->size() );

	/* ¦ξ軻(̤buff) */

	// colԤ¦ǤѲʤ
	for ( size_type r = 0 ; r < col ; ++r )
	  for ( size_type c = 0 ; c < mat->size() ; ++c )
	    buff[r][c] = (*mat)[r][c];

	for ( size_type r = col ; r < mat->size() ; ++r ) {
	  for ( size_type c = 0 ; c < mat->size() ; ++c ) {
	    for ( size_type i = 0 ; i < v.size() ; ++i )
	      buff[r][c] -= v[i] * (*mat)[i + col][c];
	    buff[r][c] *= 2 * v[r - col] / norm;
	    buff[r][c] += (*mat)[r][c];
	  }
	}

	/* ¦ξ軻(̤mat) */

	for ( size_type r = 0 ; r < mat->size() ; ++r ) {
	  for ( size_type c = 0 ; c < col ; ++c )
	    (*mat)[r][c] = buff[r][c];
	  for ( size_type c = col ; c < mat->size() ; ++c ) {
	    (*mat)[r][c] = 0;
	    for ( size_type i = 0 ; i < v.size() ; ++i )
	      (*mat)[r][c] -= v[i] * buff[r][i + col];
	    (*mat)[r][c] *= 2 * v[c - col] / norm;
	    (*mat)[r][c] += buff[r][c];
	  }
	}
      }

      /*
	Householder_TD : ϥۥˡˤ r λгѹؤѴ

	q ľ뤿ι ( H(v) )
      */
      template< class T >
      void Householder_TD
      ( SquareMatrix< T >* r, SquareMatrix< T >* q )
      {
	assert( r != nullptr );

	using size_type = typename SquareMatrix< T >::size_type;

	if ( q != nullptr ) {
	  q->assign( r->size() );
	  for ( size_type i = 0 ; i < r->size() ; ++i )
	    (*q)[i][i] = 1;
	}

	if ( r->size() < 3 ) return;

	for ( size_type c = 0 ; c < r->size() - 2 ; ++c ) {
	  // оݤ٥ȥѴ
	  auto colIt = r->column( c, c + 1 );
	  std::vector< T > v( colIt, colIt.end() );
	  T norm = CreateVec( &v );

	  // ϥۥˤѴ
	  Householder_ST( r, c + 1, v, norm );

	  // ľη׻
	  if ( q == nullptr ) continue;
	  if ( c == 0 ) {
	    for ( size_type j = 0 ; j < r->size() - 1 ; ++j ) {
	      (*q)[j + 1][j + 1] = 1 - 2 * v[j] * v[j] / norm;
	      for ( size_type i = j + 1 ; i < r->size() - 1 ; ++i )
		(*q)[j + 1][i + 1] = (*q)[i + 1][j + 1] = - 2 * v[j] * v[i] / norm;
	    }
	  } else {
	    MultHouseholder( q, c + 1, v, norm );
	  }
	}
      }

      /*
	Householder_GR : ֥󥹲žˤ뻰гѹ r ξ廰ѹ

	q ľ( G(i,j,) )sz Ͻϰ()򤽤줾ɽ
      */
      template< class T >
      void Householder_GR
      ( SquareMatrix< T >* r, SquareMatrix< T >* q, typename SquareMatrix< T >::size_type sz )
      {
	assert( r != nullptr );

	using size_type = typename SquareMatrix< T >::size_type;

	if ( q != nullptr ) {
	  q->assign( r->size() );
	  for ( size_type i = 0 ; i < r->size() ; ++i )
	    (*q)[i][i] = 1;
	}

	if ( r->size() < 2 ) return;

	for ( size_type i = 0 ; i < sz ; ++i ) {
	  // гʬβʬʤ鲿ʤ()
	  if ( (*r)[i + 1][i] == 0 ) continue;

	  T d = std::sqrt( std::pow( (*r)[i][i], 2.0 ) + std::pow( (*r)[i + 1][i], 2.0 ) );

	  T sin = (*r)[i + 1][i] / d;
	  T cos = (*r)[i][i] / d;

	  // ֥󥹲žˤ r[i+1][i]򥼥ˤ
	  for ( size_type j = 0 ; j < 2 ; ++j ) {
	    T d1 = (*r)[i][i + j];
	    T d2 = (*r)[i + 1][i + j];
	    (*r)[i][i + j] = d1 * cos + d2 * sin;
	    (*r)[i + 1][i + j] = -d1 * sin + d2 * cos;
	  }
	  if ( i + 2 < r->size() ) {
	    d = (*r)[i + 1][i + 2];
	    (*r)[i][i + 2] = d * sin;
	    (*r)[i + 1][i + 2] = d * cos;
	  }

	  // ľη׻
	  if ( q == nullptr ) continue;
	  if ( i == 0 ) {
	    (*q)[0][0] = (*q)[1][1] = cos;
	    (*q)[0][1] = sin;
	    (*q)[1][0] = -sin;
	  } else {
	    for ( size_type c = 0 ; c < r->size() ; ++c ) {
	      T d1 = (*q)[i][c] * cos + (*q)[i + 1][c] * sin;
	      T d2 = -(*q)[i][c] * sin + (*q)[i + 1][c] * cos;
	      (*q)[i][c] = d1;
	      (*q)[i + 1][c] = d2;
	    }
	  }
	}
	if ( q != nullptr )
	  q->transpose(); // ľž
      }

      /*
	Householder_EV22 : 2x2θͭͤ׻

	double a,b,c,d : (줾캸塦塦)
      */
      template< class T >
      T Householder_EV22( T a, T b, T c, T d )
      {
	T d1 = a + d;
	T d2 = a * d - b * c;

	T a1 = ( d1 + std::sqrt( d1 * d1 - 4 * d2 ) ) / 2;
	T a2 = ( d1 - std::sqrt( d1 * d1 - 4 * d2 ) ) / 2;

	if ( std::abs( d - a1 ) < std::abs( d - a2 ) )
	  return( a1 );
	else
	  return( a2 );
      }

      /*
	Householder_QR : гѹ QR Ѵˤͭͭ͡٥ȥ

	mat : Ѵоݤоι
	r : Ѵξ廰ѹǼؤΥݥ
	q : ѴľǼؤΥݥ
	e : «ȽΤΤ
	maxCnt : «ʤäκ
      */
      template< class T >
      void Householder_QR
      ( const SymmetricMatrix< T >& mat, SquareMatrix< T >* r, SquareMatrix< T >* q, T e, uint32_t maxCnt )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "EigenLib::Householder_QR";

	assert( r != nullptr );
	error_lib::NNNum( e, FUNC_NAME + " : e " + error_message::NNNUM );

	*r = mat;

	uint32_t cnt = 0; // 
	Householder_TD( r, q );
	if ( q != nullptr ) q->transpose();

	size_type sz = r->size() - 1;
	while ( cnt < maxCnt ) {
	  SquareMatrix< T > qk;

	  Householder_GR( r, &qk, sz );
	  *r *= qk; // RQ׻

	  if ( q != nullptr ) *q *= qk;   // ľη׻
	  if ( std::abs( (*r)[sz][sz - 1] ) < e ) --sz;
	  if ( sz == 0 ) break;
	  ++cnt;
	}

#ifdef DEBUG
	std::cout << FUNC_NAME + " : cnt = " << cnt + 1 << std::endl;
#endif

	// Ķ㳰ꤲ
	if ( cnt >= maxCnt )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt );
      }

      /*
	Householder_DoubleShiftQR : եդQRѴ

	mat : Ѵоݤоι
	r : Ѵξ廰ѹǼؤΥݥ
	q : ѴľǼؤΥݥ
	e : «ȽΤΤ
	maxCnt : «ʤäκ
      */
      template< class T >
      void Householder_DoubleShiftQR
      ( const SymmetricMatrix< T >& mat, SquareMatrix< T >* r, SquareMatrix< T >* q, T e, uint32_t maxCnt )
      {
	using size_type = typename SquareMatrix< T >::size_type;

	const std::string FUNC_NAME = "EigenLib::Householder_DoubleShiftQR";

	assert( r != nullptr );
	error_lib::NNNum( e, FUNC_NAME + " : e " + error_message::NNNUM );

	*r = mat;

	uint32_t cnt = 0; // 
	Householder_TD( r, q );
	if ( q != nullptr ) q->transpose();

	size_type sz = r->size() - 1;
	while ( cnt < maxCnt ) {
	  SquareMatrix< T > qk;
	  T u = Householder_EV22( (*r)[sz - 1][sz - 1], (*r)[sz - 1][sz], (*r)[sz][sz - 1], (*r)[sz][sz] );
	  for ( size_type i = 0 ; i <= sz ; ++i )
	    (*r)[i][i] -= u;

	  Householder_GR( r, &qk, sz );
	  *r *= qk; // RQ׻
	  for ( size_type i = 0 ; i <= sz ; ++i )
	    (*r)[i][i] += u;

	  if ( q != nullptr ) *q *= qk;   // ľη׻
	  if ( std::abs( (*r)[sz][sz - 1] ) < e ) --sz;
	  if ( sz == 0 ) break;
	  ++cnt;
	}

#ifdef DEBUG
	std::cout << FUNC_NAME + " : cnt = " << cnt + 1 << std::endl;
#endif

	// Ķ㳰ꤲ
	if ( cnt >= maxCnt )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt );
      }

      /**
	 ٤ˡ
      **/

      /*
	PowerIter : ٤ˡˤ a ΰܤθͭͭ͡٥ȥ

	x : ͭ٥ȥؤΥݥ(ϤǤϽͤȤѤ)
	test : «ȽѴؿ֥
	maxCnt : ȿκ
	 : ͭ
      */
      template< class T, template< class > class MATRIX, class Test >
      T PowerIter( const MATRIX< T >& a, std::vector< T >* x, Test test, uint32_t maxCnt )
      {
	const std::string FUNC_NAME = "EigenLib::PowerIter";

	assert( x != nullptr );

	error_lib::CheckMultMV( a, x->begin(), x->end(),
				FUNC_NAME + "vector x " + error_message::NEGATIVE_LENGTH,
			      FUNC_NAME + " : a.cols() " + error_message::NOT_EQUAL_LENGTH + " x->size()" );

	// θͭ٥ȥ x ǽ
	std::vector< T > newX( *x );
	if ( ! math_lib::Normalize( newX.begin(), newX.end() ) )
	  return( std::numeric_limits< T >::quiet_NaN() );

	uint32_t cnt = 0; // 
	for ( ; cnt < maxCnt ; ++cnt ) {
	  *x = newX;
	  //  x  Ax ǹ
	  matrix_lib::multMV( a, x->begin(), newX.begin() );
	  if ( ! math_lib::Normalize( newX.begin(), newX.end() ) )
	    return( std::numeric_limits< T >::quiet_NaN() );
	  // «Ƚ
	  test.init();
	  for ( typename std::vector< T >::size_type i = 0 ; i < x->size() ; ++i )
	    test( (*x)[i], newX[i] );
	  if ( test.isConvergent() )
	    break;
	}

	if ( cnt >= maxCnt )
	  throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT, maxCnt );

	matrix_lib::multMV( a, x->begin(), newX.begin() );

	return( std::inner_product( newX.begin(), newX.end(), newX.begin(), T() ) /
		std::inner_product( x->begin(), x->end(), newX.begin(), T() ) );
      }

    } // namespace eigen_lib

  } // namespace math_lib

} // namespace tadah_fussy
