/**
   @file gtk++.hpp
   @brief GTK 関連オブジェクト定義用ヘッダ

   @author tadah_fussy
   @date 2021/04/24 新規作成
**/
#ifndef GTKPP_20210424_H
#define GTKPP_20210424_H

#include <string>
#include <iostream>
#include <gtk/gtk.h>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief GTK 関連オブジェクト用名前空間
  **/
  namespace gtk
  {
    /**
       @brief ファイル選択用ダイアログ
    **/
    class FileChooser
    {
      std::string currentDir_;  // カレント・ディレクトリ
      std::string currentFile_; // 前回選択したファイル

    public:

      /// @brief デフォルト・コンストラクタ
      FileChooser() : currentDir_(), currentFile_() {}

      /// @brief ダイアログを表示してファイルを選択させる
      ///
      /// @param top トップレベルウィンドウ
      /// @param action ダイアログの用途 ( GTK_FILE_CHOOSER_ACTION_OPEN / GTK_FILE_CHOOSER_ACTION_SAVE / GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER / GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER )
      /// @param title ダイアログのタイトル
      /// @param acceptBtn 受付ボタンのラベル名
      /// @param cancelBtn キャンセルボタンのラベル名
      /// @return 応答の値 ( GTK_RESPONSE_NONE / GTK_RESPONSE_REJECT / GTK_RESPONSE_ACCEPT / GTK_RESPONSE_DELETE_EVENT )
      gint32 run( GtkWindow* top, GtkFileChooserAction action, const std::string& title, const std::string& acceptBtn, const std::string& cancelBtn );

      /// @brief 選択されたファイルの名前を返す
      ///
      /// @return 選択されたファイルの名前
      std::string file() const
      { return( currentFile_ ); }
    };
  }
}

#endif
